import pandas as pd
import numpy as np
from numpy.typing import ArrayLike

import json

from imspy.timstof.frame import TimsFrame
from imspy.timstof.slice import TimsSlice
from imspy.utility.utilities import gaussian, exp_gaussian
from imspy.simulation.isotopes import IsotopePatternGenerator, create_initial_feature_distribution
from abc import ABC, abstractmethod
from typing import Optional, Dict


class Profile:

    def __init__(self, positions: Optional[ArrayLike] = None, rel_abundances: Optional[ArrayLike] = None,
                 model_params: Optional[Dict] = None, jsons:Optional[str] = None):

        if jsons is not None:
            self._jsons = jsons
            self.positions, self.rel_abundances, self.model_params, self.access_dictionary = self._from_jsons(jsons)

        else:
            self.positions = np.asarray(positions)
            self.rel_abundances = np.asarray(rel_abundances)
            self.model_params = model_params
            self.access_dictionary = {p:i for p,i in zip(positions, rel_abundances)}
            self._jsons = self._to_jsons()

    def __iter__(self):
        self.__size = len(self.positions)
        self.__iterator_pos = 0
        return self

    def __next__(self):
        if self.__iterator_pos < self.__size:
            p = self.positions[self.__iterator_pos]
            ra = self.rel_abundances[self.__iterator_pos]
            self.__iterator_pos += 1
            return p, ra
        raise StopIteration

    def __getitem__(self, position: int):
        return self.access_dictionary[position]

    def _to_jsons(self):
        mp = {}
        for key, val in self.model_params.items():
            if isinstance(val, np.generic):
                mp[key] = val.item()
            else:
                mp[key] = val

        json_dict = {"positions": self.positions.tolist(),
                     "rel_abundances": self.rel_abundances.tolist(),
                     "model_params": mp}
        return json.dumps(json_dict, allow_nan=False)

    def _from_jsons(self, jsons:str):
        json_dict = json.loads(jsons)
        positions = json_dict["positions"]
        rel_abundances = json_dict["rel_abundances"]
        access_dictionary = {p: i for p, i in zip(positions, rel_abundances)}
        return positions, rel_abundances, json_dict["model_params"], access_dictionary

    @property
    def jsons(self):
        return self._jsons


class RTProfile(Profile):

    def __init__(self, frames: Optional[ArrayLike] = None, rel_abundances: Optional[ArrayLike] = None, model_params: Optional[Dict] = None, jsons:Optional[str] = None):
        super().__init__(frames, rel_abundances, model_params, jsons)

    @property
    def frames(self):
        return self.positions


class ScanProfile(Profile):

    def __init__(self, scans:Optional[ArrayLike] = None, rel_abundances:Optional[ArrayLike] = None, model_params: Optional[Dict] = None, jsons:Optional[str] = None):
        super().__init__(scans, rel_abundances, model_params, jsons)

    @property
    def scans(self):
        return self.positions


class ChargeProfile(Profile):

    def __init__(self, charges: Optional[ArrayLike] = None, rel_abundance: Optional[ArrayLike] = None,
                 model_params: Optional[Dict] = None, jsons: Optional[str] = None):

        abundant_charges = charges[rel_abundance > 0]

        relative_abundances_over_0 = rel_abundance[rel_abundance > 0]

        super().__init__(abundant_charges, relative_abundances_over_0, model_params, jsons)

    @property
    def charges(self):
        return self.positions


class FeatureGenerator(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def generate_feature(self, mz: float, charge: int):
        pass


class PrecursorFeatureGenerator(FeatureGenerator):

    def __init__(self):
        super(PrecursorFeatureGenerator).__init__()

    def generate_feature(self, mz: float, charge: int,
                         pattern_generator: IsotopePatternGenerator,
                         num_rt: int = 64,
                         num_im: int = 32,
                         distr_im=gaussian,
                         distr_rt=exp_gaussian,
                         rt_lower: float = -9,
                         rt_upper: float = 18,
                         im_lower: float = -4,
                         im_upper: float = 4,
                         intensity_amplitude: float = 1e3,
                         min_intensity: int = 5) -> TimsSlice:

        I = create_initial_feature_distribution(num_rt, num_im, rt_lower, rt_upper,
                                                im_lower, im_upper, distr_im, distr_rt)

        frame_list = []

        spec = pattern_generator.generate_spectrum(mz, charge, min_intensity=5, centroided=True, k=12)
        mz, intensity = spec.mz, spec.intensity / np.max(spec.intensity) * 100

        for i in range(num_rt):

            scan_arr, mz_arr, intensity_arr, tof_arr, inv_mob_arr = [], [], [], [], []

            for j in range(num_im):
                intensity_scaled = intensity * I[i, j]
                intensity_scaled = intensity_scaled * intensity_amplitude
                int_intensity = np.array([int(x) for x in intensity_scaled])

                bt = [(x, y) for x, y in zip(mz, int_intensity) if y >= min_intensity]

                mz_tmp = np.array([x for x, y in bt])
                intensity_tmp = np.array([y for x, y in bt]).astype(np.int32)

                rts = np.repeat(i, intensity_tmp.shape[0]).astype(np.float32)
                scans = np.repeat(j, intensity_tmp.shape[0]).astype(np.int32)
                tof = np.ones_like(rts).astype(np.int32)
                inv_mob = np.ones_like(scans)

                scan_arr.append(scans)
                mz_arr.append(mz_tmp)
                intensity_arr.append(intensity_tmp)
                tof_arr.append(tof)
                inv_mob_arr.append(inv_mob)

            frame = TimsFrame(None, i, float(i),
                              np.hstack(scan_arr), np.hstack(mz_arr), np.hstack(intensity_arr),
                              np.hstack(tof_arr), np.hstack(inv_mob_arr))

            frame_list.append(frame)

        return TimsSlice(None, frame_list, [])


class FeatureBatch:
    def __init__(self, feature_table: pd.DataFrame, raw_data: TimsSlice):
        """

        :param feature_table:
        :param raw_data:
        """
        self.feature_table = feature_table.sort_values(['rt_length'], ascending=False)
        self.raw_data = raw_data
        self.__feature_counter = 0
        self.current_row = self.feature_table.iloc[0]
        r = self.current_row
        self.current_feature = self.get_feature(r.mz - 0.1, r.mz + (r.num_peaks / r.charge) + 0.1,
                                                r.im_start,
                                                r.im_stop,
                                                r.rt_start,
                                                r.rt_stop)

    def __repr__(self):
        return f'FeatureBatch(features={self.feature_table.shape[0]}, slice={self.raw_data})'

    def get_feature(self, mz_min, mz_max, scan_min, scan_max, rt_min, rt_max, intensity_min=20):
        return self.raw_data.filter_ranged(mz_min=mz_min,
                                           mz_max=mz_max,
                                           scan_min=scan_min,
                                           scan_max=scan_max,
                                           intensity_min=intensity_min,
                                           rt_min=rt_min,
                                           rt_max=rt_max)

    def __iter__(self):
        return self.current_row, self.current_feature

    def __next__(self):
        feature = self.current_row
        data = self.current_feature

        if self.__feature_counter < self.feature_table.shape[0]:
            self.__feature_counter += 1
            self.current_row = self.feature_table.iloc[self.__feature_counter]
            r = self.current_row
            self.current_feature = self.get_feature(r.mz - 0.1,
                                                    r.mz + (r.num_peaks / r.charge) + 0.1,
                                                    r.im_start,
                                                    r.im_stop,
                                                    r.rt_start,
                                                    r.rt_stop)
            return feature, data

        raise StopIteration



import tensorflow as tf
import importlib.resources as resources
from imspy.utility.utilities import tokenizer_from_json
from importlib.abc import Traversable


def get_model_path(model_name: str) -> Traversable:
    """ Get the path to a pretrained model

    Args:
        model_name: The name of the model to load

    Returns:
        The path to the pretrained model
    """
    return resources.files('imspy.algorithm.pretrained').joinpath(model_name)


def load_tokenizer_from_resources(tokenizer_name: str) -> tf.keras.preprocessing.text.Tokenizer:
    """ Load a tokenizer from resources

    Returns:
        The pretrained tokenizer
    """
    return tokenizer_from_json(resources.files('imspy.algorithm.pretrained').joinpath(f'{tokenizer_name}.json'))



import numpy as np
import pandas as pd

from sagepy.core.database import PeptideIx
from sagepy.core import EnzymeBuilder, SAGE_KNOWN_MODS, validate_mods, validate_var_mods, SageSearchConfiguration
from tqdm import tqdm


class PeptideDigest:
    def __init__(self,
                 fasta_path: str,
                 missed_cleavages: int = 2,
                 min_len: int = 7,
                 max_len: int = 50,
                 cleave_at: str = 'KR',
                 restrict: str = 'P',
                 generate_decoys: bool = False,
                 c_terminal: bool = True,
                 verbose: bool = True,
                 ):

        self.verbose = verbose
        self.peptides = None
        self.min_len = min_len
        self.max_len = max_len
        self.missed_cleavages = missed_cleavages
        self.cleave_at = cleave_at
        self.restrict = restrict
        self.generate_decoys = generate_decoys
        self.c_terminal = c_terminal
        self.fasta_path = fasta_path

        self._setup()

    def _setup(self):
        try:
            fasta = open(self.fasta_path, 'r').read()
        except FileNotFoundError:
            print(f"Could not find fasta file at {self.fasta_path}")
            raise FileNotFoundError

        enzyme_builder = EnzymeBuilder(
            missed_cleavages=self.missed_cleavages,
            min_len=self.min_len,
            max_len=self.max_len,
            cleave_at=self.cleave_at,
            restrict=self.restrict,
            c_terminal=self.c_terminal
        )

        # generate static cysteine modification
        static_mods = {k: v for k, v in [SAGE_KNOWN_MODS.cysteine_static()]}

        # generate variable methionine modification
        variable_mods = {k: v for k, v in [SAGE_KNOWN_MODS.methionine_variable()]}

        # generate SAGE compatible mod representations
        static = validate_mods(static_mods)
        variable = validate_var_mods(variable_mods)

        sage_config = SageSearchConfiguration(
            fasta=fasta,
            enzyme_builder=enzyme_builder,
            static_mods=static,
            variable_mods=variable,
            generate_decoys=self.generate_decoys,
            bucket_size=int(np.power(2, 6))
        )

        indexed_db = sage_config.generate_indexed_database()

        peptide_list = []

        for i in tqdm(range(indexed_db.num_peptides), desc='Digesting peptides', disable=not self.verbose, ncols=100):
            idx = PeptideIx(idx=i)
            peptide = indexed_db[idx]

            peptide_list.append({'peptide_id': i,
                                 'sequence': peptide.to_unimod_sequence(),
                                 'proteins': '\t'.join(peptide.proteins),
                                 'decoy': peptide.decoy,
                                 'missed_cleavages': peptide.missed_cleavages,
                                 'n_term': peptide.n_term,
                                 'c_term': peptide.c_term,
                                 'monoisotopic-mass': peptide.mono_isotopic})

        self.peptides = pd.DataFrame(peptide_list)



import pandas as pd
import tensorflow as tf
import numpy as np

from abc import ABC, abstractmethod
from numpy.typing import NDArray

from tqdm import tqdm

from imspy.algorithm.utility import get_model_path
from imspy.utility import tokenize_unimod_sequence

import imspy_connector
ims = imspy_connector.py_chemistry


def load_deep_charge_state_predictor() -> tf.keras.models.Model:
    """ Get a pretrained deep predictor model

    Returns:
        The pretrained deep predictor model
    """
    return tf.keras.models.load_model(get_model_path('DeepChargeStatePredictor'))


def charge_state_distribution_from_sequence_rust(sequence: str, max_charge: int = None,
                                                 charge_probability: float = None) -> NDArray:
    return np.array(ims.simulate_charge_state_for_sequence(sequence, max_charge, charge_probability))


def charge_state_distributions_from_sequences_rust(sequences: list[str], n_threads: int = 4, max_charge: int = None,
                                                   charge_probability: float = None) -> NDArray:
    return np.array(ims.simulate_charge_states_for_sequences(sequences, n_threads, max_charge, charge_probability))


class PeptideChargeStateDistribution(ABC):
    """
    ABSTRACT INTERFACE for ionization simulation of peptides
    """

    def __init__(self):
        pass

    @abstractmethod
    def simulate_ionizations(self, sequences: list[str]) -> np.array:
        pass

    @abstractmethod
    def simulate_charge_state_distribution_pandas(self, data: pd.DataFrame) -> pd.DataFrame:
        pass


class BinomialChargeStateDistributionModel(PeptideChargeStateDistribution, ABC):

    def __init__(self,
                 charged_probability: float = .5,
                 max_charge: int = 4,
                 ):
        self.charged_probability = charged_probability
        self.max_charge = max_charge

    def simulate_ionizations(self, sequences: list[str]) -> np.array:
        return charge_state_distributions_from_sequences_rust(sequences, max_charge=self.max_charge,
                                                              charge_probability=self.charged_probability)

    def simulate_charge_state_distribution_pandas(self, data: pd.DataFrame, min_charge_contrib: float = .005) -> pd.DataFrame:
        probabilities = charge_state_distributions_from_sequences_rust(data.sequence.values, max_charge=self.max_charge,
                                                                       charge_probability=self.charged_probability)

        probabilities = probabilities / np.expand_dims(np.sum(probabilities, axis=1), axis=1)

        r_table = []

        for charges, (_, row) in tqdm(zip(probabilities, data.iterrows()), desc='flatmap charges', ncols=80,
                                      total=len(probabilities)):
            for i, charge in enumerate(charges, start=1):
                if charge >= min_charge_contrib:
                    r_table.append({'peptide_id': row.peptide_id, 'charge': i, 'relative_abundance': charge})

        return pd.DataFrame(r_table)


class DeepChargeStateDistribution(PeptideChargeStateDistribution, ABC):

    def __init__(self,
                 model: 'GRUChargeStatePredictor',
                 tokenizer: tf.keras.preprocessing.text.Tokenizer,
                 allowed_charges: NDArray = np.array([1, 2, 3, 4]),
                 name: str = 'gru_predictor',
                 verbose: bool = True
                 ):
        super(DeepChargeStateDistribution, self).__init__()
        self.model = model
        self.tokenizer = tokenizer
        self.allowed_charges = allowed_charges
        self.name = name
        self.verbose = verbose

    def _preprocess_sequences(self, sequences: list[str], pad_len: int = 50) -> np.array:
        char_tokens = [tokenize_unimod_sequence(seq) for seq in sequences]
        char_tokens = self.tokenizer.texts_to_sequences(char_tokens)
        char_tokens = tf.keras.preprocessing.sequence.pad_sequences(char_tokens, pad_len, padding='post')
        return char_tokens

    def simulate_ionizations(self, sequences: list[str], batch_size: int = 1024) -> NDArray:
        tokens = self._preprocess_sequences(sequences)
        tf_ds = tf.data.Dataset.from_tensor_slices(tokens).batch(batch_size)

        probabilities = self.model.predict(tf_ds, verbose=self.verbose)

        c_list = []

        for p in probabilities:
            c_list.append(np.random.choice(range(1, len(p) + 1), 1, p=p)[0])

        return np.array(c_list)

    def simulate_charge_state_distribution_pandas(self,
                                                  data: pd.DataFrame,
                                                  batch_size: int = 1024,
                                                  min_charge_contrib: float = .005) -> pd.DataFrame:
        """
        Simulate charge state distribution for a pandas DataFrame containing peptide sequences

        Args:
            data: pandas DataFrame containing peptide sequences
            batch_size: batch size for prediction
            min_charge_contrib: minimum relative abundance of a charge state to be included in the output

        Returns:
            pandas DataFrame containing simulated charge state distributions
        """
        tokens = self._preprocess_sequences(data.sequence.values)
        tf_ds = tf.data.Dataset.from_tensor_slices(tokens).batch(batch_size)

        probabilities = self.model.predict(tf_ds, verbose=self.verbose)

        r_table = []

        for charges, (_, row) in tqdm(zip(probabilities, data.iterrows()), desc='flatmap charges', ncols=80,
                                      total=len(probabilities)):
            for i, charge in enumerate(charges, start=1):
                if charge >= min_charge_contrib:
                    r_table.append({'peptide_id': row.peptide_id, 'charge': i, 'relative_abundance': charge})

        return pd.DataFrame(r_table)


class GRUChargeStatePredictor(tf.keras.models.Model):

    def __init__(self,
                 num_tokens,
                 max_charge=4,
                 seq_len=50,
                 emb_dim=128,
                 gru_1=128,
                 gru_2=64,
                 rdo=0.0,
                 do=0.2):
        super(GRUChargeStatePredictor, self).__init__()

        self.emb = tf.keras.layers.Embedding(input_dim=num_tokens + 1, output_dim=emb_dim, input_length=seq_len)

        self.gru1 = tf.keras.layers.Bidirectional(tf.keras.layers.GRU(gru_1, return_sequences=True,
                                                                      name='GRU1'))

        self.gru2 = tf.keras.layers.Bidirectional(tf.keras.layers.GRU(gru_2, return_sequences=False,
                                                                      name='GRU2',
                                                                      recurrent_dropout=rdo))

        self.dense1 = tf.keras.layers.Dense(128, activation='relu', name='Dense1',
                                            kernel_regularizer=tf.keras.regularizers.l1_l2(1e-3, 1e-3))

        self.dense2 = tf.keras.layers.Dense(64, activation='relu', name='Dense2',
                                            kernel_regularizer=tf.keras.regularizers.l1_l2(1e-3, 1e-3))

        self.dropout = tf.keras.layers.Dropout(do, name='Dropout')

        self.out = tf.keras.layers.Dense(max_charge, activation='softmax', name='Output')

    def call(self, inputs, **kwargs):
        """
        :param inputs: should contain: (sequence)
        """
        # get inputs
        seq = inputs
        # sequence learning
        x_recurrent = self.gru2(self.gru1(self.emb(seq)))
        # regularize
        d1 = self.dropout(self.dense1(x_recurrent))
        # output
        return self.out(self.dense2(d1))



import sqlite3
import pandas as pd
import numpy as np
from typing import List
from pathlib import Path
from imspy.simulation.utility import get_native_dataset_path
from imspy.timstof.frame import TimsFrame
from imspy.timstof import TimsDataset
import shutil


class TDFWriter:
    def __init__(
            self,
            path: str = "./",
            exp_name: str = "RAW.d",
            num_scans: int = 917,
            im_lower: float = 0.6,
            im_upper: float = 1.6,
            mz_lower: float = 100.0,
            mz_upper: float = 1700.0)\
            -> None:

        self.path = Path(path)
        self.exp_name = exp_name
        self.full_path = Path(path) / exp_name
        self.num_scans = num_scans
        self.im_lower = im_lower
        self.im_upper = im_upper
        self.mz_lower = mz_lower
        self.mz_upper = mz_upper
        self.position = 0
        self.binary_file = self.full_path / "analysis.tdf_bin"
        self.frame_meta_data = []
        self.conn = None
        self.__helper_handle = None
        self.__conn_native = None
        self._setup_connections()

    def _setup_connections(self) -> None:
        # Create the directory and connect to DB
        self.full_path.mkdir(parents=True, exist_ok=True)
        self.conn = sqlite3.connect(f'{self.full_path}/analysis.tdf')

        # generate tables for analysis.tdf DB
        tims_calib_name, tims_calib_tbl = self._create_tims_calibration_table(num_scans=self.num_scans)
        mz_calib_name, mz_calib_tbl = self._create_mz_calibration_table()
        global_meta_name, global_meta_tbl = self._create_global_meta_data_table(im_lower=self.im_lower,
                                                                                im_upper=self.im_upper,
                                                                                mz_lower=self.mz_lower,
                                                                                mz_upper=self.mz_upper)
        # Save table to analysis.tdf
        self._create_table(self.conn, mz_calib_tbl, mz_calib_name)
        self._create_table(self.conn, tims_calib_tbl, tims_calib_name)
        self._create_table(self.conn, global_meta_tbl, global_meta_name)

        # Build reference for correct translation of mz -> tof and inv_mob -> scan
        native = self.full_path / ".native/"
        native.mkdir(parents=True, exist_ok=True)
        ref_location = get_native_dataset_path()
        source_file = ref_location
        destination_file = self.full_path / ".native/"
        shutil.copytree(source_file, destination_file, dirs_exist_ok=True)
        self.__conn_native = sqlite3.connect(f'{native}/analysis.tdf')
        self._create_table(self.__conn_native, mz_calib_tbl, mz_calib_name)
        self._create_table(self.__conn_native, tims_calib_tbl, tims_calib_name)
        self._create_table(self.__conn_native, global_meta_tbl, global_meta_name)

        # setup handle for function calls
        self.__helper_handle = TimsDataset(str(native))

    @staticmethod
    def _get_table(conn, table_name: str) -> pd.DataFrame:
        # Get a table as a pandas DataFrame
        return pd.read_sql(f"SELECT * FROM {table_name}", conn)

    @staticmethod
    def _create_table(conn, table, table_name: str) -> None:
        # Create a table from a pandas DataFrame
        table.to_sql(table_name, conn, if_exists='replace', index=False)

    @staticmethod
    def _create_mz_calibration_table() -> (str, pd.DataFrame):
        col_names = ['Id', 'ModelType', 'DigitizerTimebase', 'DigitizerDelay', 'T1', 'T2',
                     'dC1', 'dC2', 'C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8',
                     'C9', 'C10', 'C11', 'C12', 'C13', 'C14']

        values = [1.0, 2.0, 0.19999999999999998, 25726.199999999997, 25.66893460672872, 25.10638483329421, 20.0, 0.0,
                  313.51167546836524, 154833.4542880268, -7.3593189552069465e-06, 313.51167546836524,
                  -7.3593189552069465e-06,
                  225.951491, 1519.712539, 7.0, 0.058519441907584555, -0.0005411344569873044, 1.8634987450927556e-06,
                  -3.1297916646308927e-09,
                  2.7471965457754794e-12, -1.207997933655641e-15, 2.095848318930801e-19]

        return "MzCalibration", pd.DataFrame([values], columns=col_names)

    @staticmethod
    def _create_tims_calibration_table(num_scans: int = 917) -> (str, pd.DataFrame):
        col_names = ['Id', 'ModelType', 'C0', 'C1', 'C2', 'C3', 'C4', 'C5', 'C6', 'C7', 'C8', 'C9']

        values = [1, 2, 1, num_scans, 218.720487393, 73.404989154, 33.027522936, 1.0, 0.042931657,
                  127.310271708, 12.676243115, 4414.816879989]

        return "TimsCalibration", pd.DataFrame([values], columns=col_names)

    @staticmethod
    def _create_global_meta_data_table(im_lower: float, im_upper: float,
                                       mz_lower: float, mz_upper) -> (str, pd.DataFrame):
        global_meta = {'SchemaType': 'TDF',
                       'SchemaVersionMajor': '3',
                       'SchemaVersionMinor': '5',
                       'AcquisitionSoftwareVendor': 'Bruker',
                       'InstrumentVendor': 'Bruker',
                       'ClosedProperly': '1',
                       'TimsCompressionType': '2',
                       'MaxNumPeaksPerScan': '1661',
                       'AnalysisId': '00000000-0000-0000-0000-000000000000',
                       'DigitizerNumSamples': '396854',
                       'MzAcqRangeLower': f'{mz_lower}',
                       'MzAcqRangeUpper': f'{mz_upper}',
                       'AcquisitionSoftware': 'timsTOF',
                       'AcquisitionSoftwareVersion': '2.0.18.0',
                       'AcquisitionFirmwareVersion': 'I4IT-9.481.28.81; ITPT-9.481.28.81; ITET-9.481.28.81; '
                                                     'FXM3-0.0.1.6; MXMC-0.0.3.4; MXIF-0.0.1.1; MXRF-0.0.1.1; '
                                                     'RFXS-0.1.3.1; RFXE-NOT_PRESENT',
                       'AcquisitionDateTime': '2021-01-15T16:15:32.327+01:00',
                       'InstrumentName': 'timsTOF Pro',
                       'InstrumentFamily': '9',
                       'InstrumentRevision': '3',
                       'InstrumentSourceType': '11',
                       'InstrumentSerialNumber': '1854399.00271',
                       'OperatorName': 'Admin',
                       'Description': 'HeLa 200ng/uL',
                       'SampleName': 'M210115_001',
                       'MethodName': '20210113_DDA PASEF-standard_1.1sec_cycletime_1600V.m',
                       'DenoisingEnabled': '0',
                       'PeakWidthEstimateValue': '0.000025',
                       'PeakWidthEstimateType': '1',
                       'PeakListIndexScaleFactor': '1',
                       'OneOverK0AcqRangeLower': f'{im_lower}',
                       'OneOverK0AcqRangeUpper': f'{im_upper}'}

        # TODO: CHECK WHAT IS GOING ON HERE
        return "GlobalMetadata", pd.DataFrame(global_meta.items(), columns=['Key', 'Value'])

    def mz_to_tof(self, mzs):
        return self.__helper_handle.mz_to_tof(1, mzs)

    def tof_to_mz(self, tofs):
        return self.__helper_handle.tof_to_mz(1, tofs)

    def inv_mobility_to_scan(self, inv_mobs):
        return self.__helper_handle.inverse_mobility_to_scan(1, inv_mobs)

    def scan_to_inv_mobility(self, scans):
        return self.__helper_handle.scan_to_inverse_mobility(1, scans)

    def __repr__(self) -> str:
        return f"TDFWriter(path={self.path}, db_name={self.exp_name}, num_scans={self.num_scans}, " \
               f"im_lower={self.im_lower}, im_upper={self.im_upper}, mz_lower={self.mz_lower}, " \
               f"mz_upper={self.mz_upper})"

    def build_frame_meta_row(self, frame: TimsFrame, scan_mode: int, frame_start_pos: int):
        r = self.__helper_handle.meta_data.iloc[0, :].copy()
        r.Id = frame.frame_id
        r.Time = frame.retention_time
        r.ScanMode = scan_mode
        r.MsMsType = frame.ms_type
        r.TimsId = frame_start_pos
        r.MaxIntensity = int(np.max(frame.intensity)) if len(frame.intensity) > 0 else 0
        r.SummedIntensities = int(np.sum(frame.intensity)) if len(frame.intensity) > 0 else 0
        r.NumScans = self.num_scans
        r.NumPeaks = len(frame.mz)

        return r

    def compress_frame(self, frame: TimsFrame) -> bytes:
        # calculate TOF using the DH of the other frame
        # TODO: move translation of mz -> tof and inv_mob -> scan to the helper handle
        tof = self.mz_to_tof(frame.mz)
        scan = self.inv_mobility_to_scan(frame.mobility)
        return self.__helper_handle.indexed_values_to_compressed_bytes(scan, tof, frame.intensity,
                                                                       total_scans=self.num_scans)

    def compress_frames(self, frames: List[TimsFrame], num_threads: int = 4) -> List[bytes]:
        return self.__helper_handle.compress_frames(frames, total_scans=self.num_scans, num_threads=num_threads)

    def write_frame(self, frame: TimsFrame, scan_mode: int) -> None:
        self.frame_meta_data.append(self.build_frame_meta_row(frame, scan_mode, self.position))
        compressed_data = self.compress_frame(frame)

        with open(self.binary_file, "ab") as bin_file:
            bin_file.write(compressed_data)
            self.position = bin_file.tell()

    def write_frames(self, frames: List[TimsFrame], scan_mode: int, num_threads: int = 4) -> None:

        # compress frames
        compressed_data = self.__helper_handle.compress_frames(
            frames,
            total_scans=self.num_scans,
            num_threads=num_threads
        )

        # write to binary file
        with open(self.binary_file, "ab") as bin_file:
            # write compressed data to binary file, and add frame meta data to list
            for frame, data in zip(frames, compressed_data):
                self.frame_meta_data.append(self.build_frame_meta_row(frame, scan_mode, self.position))
                bin_file.write(data)
                self.position = bin_file.tell()

    def get_frame_meta_data(self) -> pd.DataFrame:
        return pd.DataFrame(self.frame_meta_data)

    def write_frame_meta_data(self) -> None:
        self._create_table(self.conn, self.get_frame_meta_data(), "Frames")

    def write_dia_ms_ms_info(self, dia_ms_ms_info: pd.DataFrame) -> None:
        out = dia_ms_ms_info.rename(columns={
            'frame': 'Frame',
            'window_group': 'WindowGroup',
        })

        self._create_table(self.conn, out, "DiaFrameMsMsInfo")

    def write_dia_ms_ms_windows(self, dia_ms_ms_windows: pd.DataFrame) -> None:
        out = dia_ms_ms_windows.rename(columns={
            'window_group': 'WindowGroup',
            'scan_start': 'ScanNumBegin',
            'scan_end': 'ScanNumEnd',
            'isolation_mz': 'IsolationMz',
            'isolation_width': 'IsolationWidth',
            'collision_energy': 'CollisionEnergy',
        })

        self._create_table(self.conn, out, "DiaFrameMsMsWindows")



import imspy_connector
ims = imspy_connector.py_elements

ELEMENTAL_MONO_ISOTOPIC_MASSES = ims.get_elemental_mono_isotopic_weight_map()
ELEMENTAL_ISOTOPIC_MASSES = ims.get_elemental_isotope_weight_map()
ELEMENTAL_ISOTOPIC_ABUNDANCES = ims.get_elemental_isotope_abundance_map()



from pathlib import Path
from typing import Dict

import numpy as np
import pandas as pd
from numpy.typing import NDArray
from abc import abstractmethod, ABC

from imspy.simulation.experiment import SyntheticExperimentDataHandle
from imspy.timstof.data import AcquisitionMode
from imspy.simulation.utility import calculate_number_frames, calculate_mobility_spacing, \
    get_ms_ms_window_layout_resource_path
from imspy.simulation.tdf import TDFWriter


class TimsTofAcquisitionBuilder:
    def __init__(
            self,
            path: str,
            gradient_length: float,
            rt_cycle_length: float,
            im_lower: float,
            im_upper: float,
            mz_lower: float,
            mz_upper: float,
            num_scans: int,
            exp_name: str = "RAW.d",
    ):
        """ Base class for building TimsTOF experiments
        Parameters
        ----------
        path : str
            Path to the experiment directory
        gradient_length : float
            Length of the gradient in seconds
        rt_cycle_length : float
            Length of the RT cycle in seconds
        im_lower : float
            Lower bound of the ion mobility range (IM first scan)
        im_upper : float
            Upper bound of the ion mobility range (IM last scan)
        mz_lower : float
            Lower bound of the m/z range (m/z first scan)
        mz_upper : float
            Upper bound of the m/z range (m/z last scan)
        num_scans : int
            Number of scans that will be taken during the acquisition
        """

        self.path = path
        self.gradient_length = gradient_length
        self.rt_cycle_length = rt_cycle_length
        self.im_lower = im_lower
        self.im_upper = im_upper
        self.mz_lower = mz_lower
        self.mz_upper = mz_upper
        self.num_scans = num_scans
        self.im_cycle_length = calculate_mobility_spacing(im_lower, im_upper, num_scans)
        self.num_frames = calculate_number_frames(gradient_length, rt_cycle_length)
        self.num_scans = num_scans
        # Create the TDFWriter, used to deal with bruker binary format writing and metadata for libtimsdata.so
        self.tdf_writer = TDFWriter(
            path=self.path,
            exp_name=exp_name,
            num_scans=self.num_scans,
            im_lower=self.im_lower,
            im_upper=self.im_upper,
            mz_lower=self.mz_lower,
            mz_upper=self.mz_upper
        )
        # Create the SyntheticExperimentDataHandle, which is used to deal with the sqlite database of synthetic data
        self.synthetics_handle = SyntheticExperimentDataHandle(database_path=self.path)
        self.frame_table = None
        self.scan_table = None

    def generate_frame_table(self, verbose: bool = True) -> pd.DataFrame:
        if verbose:
            print('Generating frame layout.')
        frames = []
        for i in range(self.num_frames):
            frame_id = i + 1
            time = frame_id * self.rt_cycle_length
            frames.append({'frame_id': frame_id, 'time': time})

        return pd.DataFrame(frames)

    def generate_scan_table(self, verbose: bool = True) -> pd.DataFrame:
        if verbose:
            print('Generating scan layout.')

        scans = np.arange(self.num_scans)[::-1]
        mobilities = self.tdf_writer.scan_to_inv_mobility(scans)

        return pd.DataFrame({'scan': scans, 'mobility': mobilities})

    def __repr__(self):
        return (f"TimsTofAcquisitionBuilder(path={self.path}, gradient_length={np.round(self.gradient_length / 60)} "
                f"min, mobility_range: {self.im_lower}-{self.im_upper}, "
                f"num_frames: {self.num_frames}, num_scans: {self.num_scans})")

    @abstractmethod
    def calculate_frame_types(self, *args) -> NDArray:
        pass


class TimsTofAcquisitionBuilderDDA(TimsTofAcquisitionBuilder, ABC):
    def __init__(self,
                 path: str,
                 verbose: bool = True,
                 precursor_every: int = 7,
                 gradient_length=120 * 60,
                 rt_cycle_length=0.109,
                 im_lower=0.6,
                 im_upper=1.6,
                 num_scans=917,
                 mz_lower: float = 150,
                 mz_upper: float = 1700,
                 exp_name: str = "RAW.d"
                 ):
        super().__init__(path, gradient_length, rt_cycle_length, im_lower, im_upper, mz_lower, mz_upper, num_scans, exp_name=exp_name)
        self.scan_table = None
        self.frame_table = None
        self.precursor_every = precursor_every
        self.acquisition_mode = AcquisitionMode('DDA')
        self.verbose = verbose

        self._setup(verbose=verbose)

    def calculate_frame_types(self, table: pd.DataFrame, precursor_every: int = 7, verbose: bool = True) -> NDArray:
        if verbose:
            print(f'Calculating frame types, precursor frame will be taken every {precursor_every} rt cycles.')
        return np.array([0 if (x - 1) % (precursor_every + 1) == 0 else 8 for x in table.frame_id])

    def _setup(self, verbose: bool = True):
        self.frame_table = self.generate_frame_table(verbose=verbose)
        self.scan_table = self.generate_scan_table(verbose=verbose)
        self.frame_table['ms_type'] = self.calculate_frame_types(table=self.frame_table,
                                                                 precursor_every=self.precursor_every, verbose=verbose)

        self.synthetics_handle.create_table(
            table_name='frames',
            table=self.frame_table
        )
        self.synthetics_handle.create_table(
            table_name='scans',
            table=self.scan_table
        )


class TimsTofAcquisitionBuilderDIA(TimsTofAcquisitionBuilder, ABC):
    def __init__(self,
                 path: str,
                 window_group_file: str,
                 acquisition_name: str = "dia",
                 exp_name: str = "RAW",
                 verbose: bool = True,
                 precursor_every: int = 16,
                 gradient_length=50 * 60,
                 rt_cycle_length=0.1054,
                 im_lower=0.6,
                 im_upper=1.5,
                 num_scans=927,
                 mz_lower: float = 100,
                 mz_upper: float = 1700,
                 ):

        super().__init__(path, gradient_length, rt_cycle_length, im_lower, im_upper, mz_lower, mz_upper, num_scans,
                         exp_name=exp_name)

        self.acquisition_name = acquisition_name
        self.scan_table = None
        self.frame_table = None
        self.frames_to_window_groups = None
        self.dia_ms_ms_windows = pd.read_csv(window_group_file)

        # check if the number of scans in the window group file matches the number of scans in the experiment
        last_scan_in_table = self.dia_ms_ms_windows.iloc[-1].scan_end
        assert num_scans == last_scan_in_table, f"Number of scans in the window group file ({last_scan_in_table}) " \
                                                f"does not match the number of scans in the experiment ({num_scans})"

        self.acquisition_mode = AcquisitionMode('DIA')
        self.verbose = verbose
        self.precursor_every = precursor_every

        self._setup(verbose=verbose)

    def calculate_frame_types(self, verbose: bool = True) -> NDArray:
        if verbose:
            print(f'Calculating frame types, precursor frame will be taken every {self.precursor_every + 1} rt cycles.')
        return np.array([0 if (x - 1) % (self.precursor_every + 1) == 0 else 9 for x in self.frame_table.frame_id])

    def generate_frame_to_window_group_table(self, precursors_every: int = 16, verbose: bool = True) -> pd.DataFrame:
        if verbose:
            print(f'generating frame to window group table.')

        table_list = []
        frame_ids = self.frame_table[self.frame_table.ms_type > 0].frame_id.values
        for i, frame_id in enumerate(frame_ids):
            wg = i % precursors_every + 1
            table_list.append({'frame': frame_id, 'window_group': wg})

        return pd.DataFrame(table_list)

    def _setup(self, verbose: bool = True):
        self.frame_table = self.generate_frame_table(verbose=verbose)
        self.scan_table = self.generate_scan_table(verbose=verbose)
        self.frame_table['ms_type'] = self.calculate_frame_types(verbose=verbose)
        self.frames_to_window_groups = self.generate_frame_to_window_group_table(precursors_every=self.precursor_every)

        self.synthetics_handle.create_table(
            table_name='frames',
            table=self.frame_table
        )
        self.synthetics_handle.create_table(
            table_name='scans',
            table=self.scan_table
        )
        self.synthetics_handle.create_table(
            table_name='dia_ms_ms_info',
            table=self.frames_to_window_groups
        )
        self.synthetics_handle.create_table(
            table_name='dia_ms_ms_windows',
            table=self.dia_ms_ms_windows
        )

    @staticmethod
    def from_config(
            path: str,
            exp_name: str,
            config: Dict[str, any],
            verbose: bool = True
    ) -> 'TimsTofAcquisitionBuilderDIA':

        acquisition_name = config['name'].lower().replace('pasef', '')
        window_group_file = get_ms_ms_window_layout_resource_path(acquisition_name)

        return TimsTofAcquisitionBuilderDIA(
            path=Path(path) / exp_name,
            window_group_file=window_group_file,
            exp_name=exp_name + ".d",
            verbose=verbose,
            acquisition_name=acquisition_name,
            precursor_every=config['precursor_every'],
            gradient_length=config['gradient_length'],
            rt_cycle_length=config['rt_cycle_length'],
            im_lower=config['im_lower'],
            im_upper=config['im_upper'],
            num_scans=config['num_scans'],
            mz_lower=config['mz_lower'],
            mz_upper=config['mz_upper'],
        )

    def __repr__(self):
        return (f"TimsTofAcquisitionBuilderDIA(name={self.name}, path={self.path}, gradient_length={np.round(self.gradient_length / 60)} "
                f"min, mobility_range: {self.im_lower}-{self.im_upper}, "
                f"num_frames: {self.num_frames}, num_scans: {self.num_scans})")



import io
import json
import math
import numba
import numpy as np
import tensorflow as tf
from typing import List, Optional

from numpy.typing import ArrayLike


@numba.jit(nopython=True)
def normal_pdf(x: ArrayLike, mass: float, s: float = 0.001, inv_sqrt_2pi: float = 0.3989422804014327, normalize: bool = False):
    """
    :param inv_sqrt_2pi:
    :param x:
    :param mass:
    :param s:
    :return:
    """
    a = (x - mass) / s
    if normalize:
        return np.exp(-0.5 * np.power(a, 2))
    else:
        return inv_sqrt_2pi / s * np.exp(-0.5 * np.power(a, 2))


@numba.jit(nopython=True)
def gaussian(x, μ=0, σ=1):
    """
    Gaussian function
    :param x:
    :param μ:
    :param σ:
    :return:
    """
    A = 1 / np.sqrt(2 * np.pi * np.power(σ, 2))
    B = np.exp(- (np.power(x - μ, 2) / 2 * np.power(σ, 2)))

    return A * B


@numba.jit(nopython=True)
def exp_distribution(x, λ=1):
    """
    Exponential function
    :param x:
    :param λ:
    :return:
    """
    if x > 0:
        return λ * np.exp(-λ * x)
    return 0


@numba.jit(nopython=True)
def exp_gaussian(x, μ=-3, σ=1, λ=.25):
    """
    laplacian distribution with exponential decay
    :param x:
    :param μ:
    :param σ:
    :param λ:
    :return:
    """
    A = λ / 2 * np.exp(λ / 2 * (2 * μ + λ * np.power(σ, 2) - 2 * x))
    B = math.erfc((μ + λ * np.power(σ, 2) - x) / (np.sqrt(2) * σ))
    return A * B


class NormalDistribution:
    def __init__(self, μ: float, σ: float):
        self.μ = μ
        self.σ = σ

    def __call__(self, x):
        return gaussian(x, self.μ, self.σ)


class ExponentialGaussianDistribution:
    def __init__(self, μ: float = -3, σ: float = 1, λ: float = .25):
        self.μ = μ
        self.σ = σ
        self.λ = λ

    def __call__(self, x):
        return exp_gaussian(x, self.μ, self.σ, self.λ)



class TokenSequence:

    def __init__(self, sequence_tokenized: Optional[List[str]] = None, jsons:Optional[str] = None):
        if jsons is not None:
            self.sequence_tokenized = self._from_jsons(jsons)
            self._jsons = jsons
        else:
            self.sequence_tokenized = sequence_tokenized
            self._jsons = self._to_jsons()

    def _to_jsons(self):
        json_dict = self.sequence_tokenized
        return json.dumps(json_dict)

    def _from_jsons(self, jsons:str):
        return json.loads(jsons)

    @property
    def jsons(self):
        return self._jsons


def is_unimod_start(char:str):
    """
    Tests if char is start of unimod
    bracket

    :param char: Character of a proForma formatted aa sequence
    :type char: str
    :return: Wether char is start of unimod bracket
    :rtype: bool
    """
    if char in ["(","[","{"]:
        return True
    else:
        return False


def is_unimod_end(char:str):
    """
    Tests if char is end of unimod
    bracket

    :param char: Character of a proForma formatted aa sequence
    :type char: str
    :return: Wether char is end of unimod bracket
    :rtype: bool
    """
    if char in [")","]","}"]:
        return True
    else:
        return False


def tokenize_proforma_sequence(sequence: str):
    """
    Tokenize a ProForma formatted sequence string.

    :param sequence: Sequence string (ProForma formatted)
    :type sequence: str
    :return: List of tokens
    :rtype: List
    """
    sequence = sequence.upper().replace("(","[").replace(")","]")
    token_list = ["<START>"]
    in_unimod_bracket = False
    tmp_token = ""

    for aa in sequence:
        if is_unimod_start(aa):
            in_unimod_bracket = True
        if in_unimod_bracket:
            if is_unimod_end(aa):
                in_unimod_bracket = False
            tmp_token += aa
            continue
        if tmp_token != "":
            token_list.append(tmp_token)
            tmp_token = ""
        tmp_token += aa

    if tmp_token != "":
        token_list.append(tmp_token)

    if len(token_list) > 1:
        if token_list[1].find("UNIMOD:1") != -1:
            token_list[1] = "<START>"+token_list[1]
            token_list = token_list[1:]
    token_list.append("<END>")

    return token_list


def get_aa_num_proforma_sequence(sequence:str):
    """
    get number of amino acids in sequence

    :param sequence: proforma formatted aa sequence
    :type sequence: str
    :return: Number of amino acids
    :rtype: int
    """
    num_aa = 0
    inside_bracket = False

    for aa in sequence:
        if is_unimod_start(aa):
            inside_bracket = True
        if inside_bracket:
            if is_unimod_end(aa):
                inside_bracket = False
            continue
        num_aa += 1
    return num_aa


def re_index_indices(ids):
    """Re-index indices, i.e. replace gaps in indices with consecutive numbers.
    Can be used, e.g., to re-index frame IDs from precursors for visualization.
    Args:
        ids: Indices.
    Returns:
        Indices.
    """
    _, inverse = np.unique(ids, return_inverse=True)
    return inverse


def tokenizer_to_json(tokenizer: tf.keras.preprocessing.text.Tokenizer, path: str):
    """
    save a fit keras tokenizer to json for later use
    :param tokenizer: fit keras tokenizer to save
    :param path: path to save json to
    """
    tokenizer_json = tokenizer.to_json()
    with io.open(path, 'w', encoding='utf-8') as f:
        f.write(json.dumps(tokenizer_json, ensure_ascii=False))


def tokenizer_from_json(path: str):
    """
    load a pre-fit tokenizer from a json file
    :param path: path to tokenizer as json file
    :return: a keras tokenizer loaded from json
    """
    with open(path) as f:
        data = json.load(f)
    return tf.keras.preprocessing.text.tokenizer_from_json(data)



import pandas as pd

from typing import List
from numpy.typing import NDArray

from tensorflow import sparse as sp

import numpy as np
from imspy.data.spectrum import TimsSpectrum, IndexedMzSpectrum
from imspy.utility.utilities import re_index_indices

import imspy_connector
ims = imspy_connector.py_tims_frame


class TimsFrame:
    def __init__(self, frame_id: int, ms_type: int, retention_time: float, scan: NDArray[np.int32],
                 mobility: NDArray[np.float64], tof: NDArray[np.int32],
                 mz: NDArray[np.float64], intensity: NDArray[np.float64]):
        """TimsFrame class.

        Args:
            frame_id (int): Frame ID.
            ms_type (int): MS type.
            retention_time (float): Retention time.
            scan (NDArray[np.int32]): Scan.
            mobility (NDArray[np.float64]): Inverse mobility.
            tof (NDArray[np.int32]): Time of flight.
            mz (NDArray[np.float64]): m/z.
            intensity (NDArray[np.float64]): Intensity.

        Raises:
            AssertionError: If the length of the scan, mobility, tof, mz and intensity arrays are not equal.
        """

        assert len(scan) == len(mobility) == len(tof) == len(mz) == len(intensity), \
            "The length of the scan, mobility, tof, mz and intensity arrays must be equal."

        self.__frame_ptr = ims.PyTimsFrame(frame_id, ms_type, retention_time, scan, mobility, tof, mz, intensity)

    def __add__(self, other: 'TimsFrame') -> 'TimsFrame':
        """Add two TimsFrames together.

        Args:
            other (TimsFrame): TimsFrame to add.

        Returns:
            TimsFrame: Sum of the two TimsFrames.
        """
        return TimsFrame.from_py_tims_frame(self.__frame_ptr + other.__frame_ptr)

    @classmethod
    def from_py_tims_frame(cls, frame: ims.PyTimsFrame):
        """Create a TimsFrame from a PyTimsFrame.

        Args:
            frame (pims.PyTimsFrame): PyTimsFrame to create the TimsFrame from.

        Returns:
            TimsFrame: TimsFrame created from the PyTimsFrame.
        """
        instance = cls.__new__(cls)
        instance.__frame_ptr = frame
        return instance

    @property
    def frame_id(self) -> int:
        """Frame ID.

        Returns:
            int: Frame ID.
        """
        return self.__frame_ptr.frame_id

    @property
    def ms_type_as_string(self) -> str:
        """MS type.

        Returns:
            int: MS type.
        """
        return self.__frame_ptr.ms_type

    @property
    def ms_type(self) -> int:
        """MS type.

        Returns:
            int: MS type.
        """
        return self.__frame_ptr.ms_type_numeric

    @property
    def retention_time(self) -> float:
        """Retention time.

        Returns:
            float: Retention time.
        """
        return self.__frame_ptr.retention_time

    @property
    def scan(self) -> NDArray[np.int32]:
        """Scan.

        Returns:
            NDArray[np.int32]: Scan.
        """
        return self.__frame_ptr.scan

    @property
    def mobility(self) -> NDArray[np.float64]:
        """Inverse mobility.

        Returns:
            NDArray[np.float64]: Inverse mobility.
        """
        return self.__frame_ptr.mobility

    @property
    def tof(self) -> NDArray[np.int32]:
        """Time of flight.

        Returns:
            NDArray[np.int32]: Time of flight.
        """
        return self.__frame_ptr.tof

    @tof.setter
    def tof(self, tof: NDArray[np.int32]):
        self.__frame_ptr.tof = tof

    @property
    def mz(self) -> NDArray[np.float64]:
        """m/z.

        Returns:
            NDArray[np.float64]: m/z.
        """
        return self.__frame_ptr.mz

    @property
    def intensity(self) -> NDArray[np.float64]:
        """Intensity.

        Returns:
            NDArray[np.float64]: Intensity.
        """
        return self.__frame_ptr.intensity

    @property
    def df(self) -> pd.DataFrame:
        """ Data as a pandas DataFrame.

        Returns:
            pd.DataFrame: Data.
        """

        return pd.DataFrame({
            'frame': np.repeat(self.frame_id, len(self.scan)),
            'retention_time': np.repeat(self.retention_time, len(self.scan)),
            'scan': self.scan,
            'mobility': self.mobility,
            'tof': self.tof,
            'mz': self.mz,
            'intensity': self.intensity})

    def filter(self,
               mz_min: float = 0.0,
               mz_max: float = 2000.0,
               scan_min: int = 0,
               scan_max: int = 1000,
               mobility_min: float = 0.0,
               mobility_max: float = 2.0,
               intensity_min: float = 0.0,
               intensity_max: float = 1e9,
               ) -> 'TimsFrame':
        """Filter the frame for a given m/z range, scan range and intensity range.

        Args:
            mz_min (float): Minimum m/z value.
            mz_max (float): Maximum m/z value.
            scan_min (int, optional): Minimum scan value. Defaults to 0.
            scan_max (int, optional): Maximum scan value. Defaults to 1000.
            mobility_min (float, optional): Minimum inverse mobility value. Defaults to 0.0.
            mobility_max (float, optional): Maximum inverse mobility value. Defaults to 2.0.
            intensity_min (float, optional): Minimum intensity value. Defaults to 0.0.
            intensity_max (float, optional): Maximum intensity value. Defaults to 1e9.

        Returns:
            TimsFrame: Filtered frame.
        """

        return TimsFrame.from_py_tims_frame(self.__frame_ptr.filter_ranged(mz_min, mz_max, scan_min, scan_max, mobility_min, mobility_max,
                                                                           intensity_min, intensity_max))

    def to_indexed_mz_spectrum(self) -> 'IndexedMzSpectrum':
        """Convert the frame to an IndexedMzSpectrum.

        Returns:
            IndexedMzSpectrum: IndexedMzSpectrum.
        """
        return IndexedMzSpectrum.from_py_indexed_mz_spectrum(self.__frame_ptr.to_indexed_mz_spectrum())

    def to_resolution(self, resolution: int) -> 'TimsFrame':
        """Convert the frame to a given resolution.

        Args:
            resolution (int): Resolution.

        Returns:
            TimsFrame: Frame with the given resolution.
        """
        return TimsFrame.from_py_tims_frame(self.__frame_ptr.to_resolution(resolution))

    def vectorized(self, resolution: int = 2) -> 'TimsFrameVectorized':
        """Convert the frame to a vectorized frame.

        Args:
            resolution (int, optional): Resolution. Defaults to 2.

        Returns:
            TimsFrameVectorized: Vectorized frame.
        """
        return TimsFrameVectorized.from_py_tims_frame_vectorized(self.__frame_ptr.vectorized(resolution))

    def to_tims_spectra(self) -> List['TimsSpectrum']:
        """Convert the frame to a list of TimsSpectrum.

        Returns:
            List[TimsSpectrum]: List of TimsSpectrum.
        """
        return [TimsSpectrum.from_py_tims_spectrum(spec) for spec in self.__frame_ptr.to_tims_spectra()]

    def to_windows(self, window_length: float = 10, overlapping: bool = True, min_num_peaks: int = 5,
                   min_intensity: float = 1) -> List[TimsSpectrum]:
        """Convert the frame to a list of windows.

        Args:
            window_length (float, optional): Window length. Defaults to 10.
            overlapping (bool, optional): Whether the windows should overlap. Defaults to True.
            min_num_peaks (int, optional): Minimum number of peaks in a window. Defaults to 5.
            min_intensity (float, optional): Minimum intensity of a peak in a window. Defaults to 1.

        Returns:
            List[MzSpectrum]: List of windows.
        """
        return [TimsSpectrum.from_py_tims_spectrum(spec) for spec in self.__frame_ptr.to_windows(
            window_length, overlapping, min_num_peaks, min_intensity)]

    @classmethod
    def from_windows(cls, windows: List[TimsSpectrum]) -> 'TimsFrame':
        """Create a TimsFrame from a list of windows.

        Args:
            windows (List[TimsSpectrum]): List of windows.

        Returns:
            TimsFrame: TimsFrame created from the windows.
        """
        return TimsFrame.from_py_tims_frame(ims.PyTimsFrame.from_windows(
            [spec.get_spec_ptr() for spec in windows]
        ))

    @classmethod
    def from_tims_spectra(cls, spectra: List[TimsSpectrum]) -> 'TimsFrame':
        """Create a TimsFrame from a list of TimsSpectrum.

        Args:
            spectra (List[TimsSpectrum]): List of TimsSpectrum.

        Returns:
            TimsFrame: TimsFrame created from the TimsSpectrum.
        """
        return TimsFrame.from_py_tims_frame(ims.PyTimsFrame.from_tims_spectra(
            [spec.get_spec_ptr() for spec in spectra]
        ))

    def to_dense_windows(self, window_length: float = 10, resolution: int = 1, overlapping: bool = True,
                         min_num_peaks: int = 5, min_intensity: float = 0.0) -> NDArray[np.float64]:

        rows, cols, values, scans, window_indices = self.__frame_ptr.to_dense_windows(window_length, resolution,
                                                                                      overlapping, min_num_peaks,
                                                                                      min_intensity)

        return scans, window_indices, np.reshape(values, (rows, cols))

    def get_frame_ptr(self):
        return self.__frame_ptr

    def __repr__(self):
        return (f"TimsFrame(frame_id={self.__frame_ptr.frame_id}, ms_type={self.__frame_ptr.ms_type}, "
                f"num_peaks={len(self.__frame_ptr.mz)})")


class TimsFrameVectorized:
    def __init__(self, frame_id: int, ms_type: int, retention_time: float, scan: NDArray[np.int32],
                 mobility: NDArray[np.float64], tof: NDArray[np.int32],
                 indices: NDArray[np.int32], intensity: NDArray[np.float64]):
        """TimsFrameVectorized class.

        Args:
            frame_id (int): Frame ID.
            ms_type (int): MS type.
            retention_time (float): Retention time.
            scan (NDArray[np.int32]): Scan.
            mobility (NDArray[np.float64]): Inverse mobility.
            tof (NDArray[np.int32]): Time of flight.
            indices (NDArray[np.int32]): Indices.
            intensity (NDArray[np.float64]): Intensity.

        Raises:
            AssertionError: If the length of the scan, mobility, tof, indices and intensity arrays are not equal.
        """

        assert len(scan) == len(mobility) == len(tof) == len(indices) == len(intensity), \
            "The length of the scan, mobility, tof, indices and intensity arrays must be equal."

        self.__frame_ptr = ims.PyTimsFrameVectorized(frame_id, ms_type, retention_time, scan, mobility, tof, indices,
                                                      intensity)

    @classmethod
    def from_py_tims_frame_vectorized(cls, frame: ims.PyTimsFrameVectorized):
        """Create a TimsFrameVectorized from a PyTimsFrameVectorized.

        Args:
            frame (pims.PyTimsFrameVectorized): PyTimsFrameVectorized to create the TimsFrameVectorized from.

        Returns:
            TimsFrameVectorized: TimsFrameVectorized created from the PyTimsFrameVectorized.
        """
        instance = cls.__new__(cls)
        instance.__frame_ptr = frame
        return instance

    @property
    def frame_id(self) -> int:
        """Frame ID.

        Returns:
            int: Frame ID.
        """
        return self.__frame_ptr.frame_id

    @property
    def ms_type(self) -> str:
        """MS type.

        Returns:
            int: MS type.
        """
        return self.__frame_ptr.ms_type_as_string

    @property
    def retention_time(self) -> float:
        """Retention time.

        Returns:
            float: Retention time.
        """
        return self.__frame_ptr.retention_time

    @property
    def scan(self) -> NDArray[np.int32]:
        """Scan.

        Returns:
            NDArray[np.int32]: Scan.
        """
        return self.__frame_ptr.scan

    @property
    def mobility(self) -> NDArray[np.float64]:
        """Inverse mobility.

        Returns:
            NDArray[np.float64]: Inverse mobility.
        """
        return self.__frame_ptr.mobility

    @property
    def tof(self) -> NDArray[np.int32]:
        """Time of flight.

        Returns:
            NDArray[np.int32]: Time of flight.
        """
        return self.__frame_ptr.tof

    @property
    def indices(self) -> NDArray[np.int32]:
        """Indices.

        Returns:
            NDArray[np.int32]: Indices.
        """
        return self.__frame_ptr.indices

    @property
    def intensity(self) -> NDArray[np.float64]:
        """Intensity.

        Returns:
            NDArray[np.float64]: Intensity.
        """
        return self.__frame_ptr.values

    @property
    def df(self) -> pd.DataFrame:
        """ Data as a pandas DataFrame.

        Returns:
            pd.DataFrame: Data.
        """

        return pd.DataFrame({
            'frame': np.repeat(self.frame_id, len(self.scan)),
            'retention_time': np.repeat(self.retention_time, len(self.scan)),
            'scan': self.scan,
            'mobility': self.mobility,
            'tof': self.tof,
            'indices': self.indices,
            'intensity': self.intensity})

    def __repr__(self):
        return (f"TimsFrameVectorized(frame_id={self.__frame_ptr.frame_id}, ms_type={self.__frame_ptr.ms_type}, "
                f"num_peaks={len(self.__frame_ptr.indices)})")

    def get_tensor_repr(self, dense=True, zero_indexed=True, re_index=True, scan_max=None, index_max=None):
        s = self.scan
        f = self.indices
        i = self.intensity

        if zero_indexed:
            s = s - np.min(s)
            f = f - np.min(f)

        if re_index:
            f = re_index_indices(f)

        if scan_max is None:
            m_s = np.max(s) + 1
        else:
            m_s = scan_max + 1

        if index_max is None:
            m_f = np.max(f) + 1
        else:
            m_f = index_max + 1

        sv = sp.reorder(sp.SparseTensor(indices=np.c_[s, f], values=i, dense_shape=(m_s, m_f)))

        if dense:
            return sp.to_dense(sv)
        else:
            return sv



from __future__ import annotations
import json
import numpy as np
from typing import List, Tuple, Callable
import pandas as pd
from numpy.typing import NDArray
from scipy.signal import find_peaks
import imspy_connector
ims = imspy_connector.py_spectrum


def get_peak_integral(peaks: NDArray[np.int32], peak_info: dict) -> NDArray[np.float64]:
    """Calculates the integral of the peaks in a spectrum.

    Args:
        peaks (NDArray[np.int32]): Peak indices.
        peak_info (dict): Peak info.

    Returns:
        NDArray[np.float64]: Peak integrals.
    """
    integrals = np.zeros(len(peaks), dtype=np.float64)
    FWHM = peak_info['widths']
    h = peak_info['prominences']
    integrals = np.sqrt(2*np.pi) * h * FWHM / (2*np.sqrt(2*np.log(2)))
    return integrals


class IndexedMzSpectrum:
    def __init__(self, index: NDArray[np.int32], mz: NDArray[np.float64], intensity: NDArray[np.float64]):
        """IndexedMzSpectrum class.

        Args:
            index (NDArray[np.int32]): Index.
            mz (NDArray[np.float64]): m/z.
            intensity (NDArray[np.float64]): Intensity.

        Raises:
            AssertionError: If the length of the index, mz and intensity arrays are not equal.
        """
        assert len(index) == len(mz) == len(intensity), ("The length of the index, mz and intensity arrays must be "
                                                         "equal.")
        self.__spec_ptr = ims.PyIndexedMzSpectrum(index, mz, intensity)

    @classmethod
    def from_py_indexed_mz_spectrum(cls, spec: ims.PyIndexedMzSpectrum):
        """Create a IndexedMzSpectrum from a PyIndexedMzSpectrum.

        Args:
            spec (pims.PyIndexedMzSpectrum): PyIndexedMzSpectrum to create the IndexedMzSpectrum from.

        Returns:
            IndexedMzSpectrum: IndexedMzSpectrum created from the PyIndexedMzSpectrum.
        """
        instance = cls.__new__(cls)
        instance.__spec_ptr = spec
        return instance

    @property
    def index(self) -> NDArray[np.int32]:
        """Index.

        Returns:
            NDArray[np.int32]: Index.
        """
        return self.__spec_ptr.index

    @property
    def mz(self) -> NDArray[np.float64]:
        """m/z.

        Returns:
            NDArray[np.float64]: m/z.
        """
        return self.__spec_ptr.mz

    @property
    def intensity(self) -> NDArray[np.float64]:
        """Intensity.

        Returns:
            NDArray[np.float64]: Intensity.
        """
        return self.__spec_ptr.intensity

    def filter(self, mz_min: float = 0.0, mz_max: float = 2000.0, intensity_min: float = 0.0,
               intensity_max: float = 1e9) -> 'IndexedMzSpectrum':
        """Filter the spectrum for a given m/z range and intensity range.

        Args:
            mz_min (float): Minimum m/z value.
            mz_max (float): Maximum m/z value.
            intensity_min (float, optional): Minimum intensity value. Defaults to 0.0.
            intensity_max (float, optional): Maximum intensity value. Defaults to 1e9.

        Returns:
            IndexedMzSpectrum: Filtered spectrum.
        """
        return IndexedMzSpectrum.from_py_indexed_mz_spectrum(
            self.__spec_ptr.filter_ranged(mz_min, mz_max, intensity_min, intensity_max))

    @property
    def df(self) -> pd.DataFrame:
        """Data.

        Returns:
            pd.DataFrame: Data.
        """

        return pd.DataFrame({'index': self.index, 'mz': self.mz, 'intensity': self.intensity})

    def get_spec_ptr(self) -> ims.PyIndexedMzSpectrum:
        """Get the spec_ptr.

        Returns:
            pims.PyIndexedMzSpectrum: spec_ptr.
        """
        return self.__spec_ptr

    def __repr__(self):
        return f"IndexedMzSpectrum(num_peaks={len(self.index)})"


class MzSpectrum:
    
    @classmethod
    def from_jsons(cls, jsons: str) -> MzSpectrum:
        json_dict:dict = json.loads(jsons)
        mz = json_dict["mz"]
        intensity = json_dict["intensity"]
        return cls(np.array(mz, dtype=np.float64), np.array(intensity, dtype=np.float64))
    
    @classmethod
    def from_mz_spectra_list(cls, spectra_list:List[MzSpectrum], resolution: int) -> MzSpectrum:
        """Generates a convoluted mass spectrum by adding all spectra in the given list.

        Args:
            spectra_list (List[MzSpectrum]): List of mass spectra.
            resolution (int): Desired resolution of returned spectrum.

        Returns:
            MzSpectrum: Convoluted spectrum.
        """
        return cls.from_py_mz_spectrum(ims.PyMzSpectrum.from_mzspectra_list([spectrum.__spec_ptr for spectrum in spectra_list], resolution))
    
    def __init__(self, mz: NDArray[np.float64], intensity: NDArray[np.float64]):
        """MzSpectrum class.

        Args:
            mz (NDArray[np.float64]): m/z.
            intensity (NDArray[np.float64]): Intensity.

        Raises:
            AssertionError: If the length of the mz and intensity arrays are not equal.
        """
        assert len(mz) == len(intensity), "The length of the mz and intensity arrays must be equal."
        self.__spec_ptr = ims.PyMzSpectrum(mz, intensity)

    @property
    def mz(self) -> NDArray[np.float64]:
        """m/z.

        Returns:
            NDArray[np.float64]: m/z.
        """
        return self.__spec_ptr.mz

    @property
    def intensity(self) -> NDArray[np.float64]:
        """Intensity.

        Returns:
            NDArray[np.float64]: Intensity.
        """
        return self.__spec_ptr.intensity

    @property
    def df(self) -> pd.DataFrame:
        """Data.

        Returns:
            pd.DataFrame: Data.
        """

        return pd.DataFrame({'mz': self.mz, 'intensity': self.intensity})

    @classmethod
    def from_py_mz_spectrum(cls, spec: ims.PyMzSpectrum):
        """Create a MzSpectrum from a PyMzSpectrum.

        Args:
            spec (pims.PyMzSpectrum): PyMzSpectrum to create the MzSpectrum from.

        Returns:
            MzSpectrum: MzSpectrum created from the PyMzSpectrum.
        """
        instance = cls.__new__(cls)
        instance.__spec_ptr = spec
        return instance

    def __repr__(self):
        return f"MzSpectrum(num_peaks={len(self.mz)})"

    def __add__(self, other: MzSpectrum) -> MzSpectrum:
        """Overwrite + operator for adding of spectra

        Args:
            other (MzSpectrum): Other spectrum.

        Returns:
            MzSpectrum: Sum of spectra
        """
        return self.from_py_mz_spectrum(self.__spec_ptr + other.__spec_ptr)
    
    def __mul__(self, scale) -> MzSpectrum:
        """Overwrite * operator for scaling of spectrum

        Args:
            scale (float): Scale.

        Returns:
            MzSpectrum: Scaled spectrum
        """
        return self.from_py_mz_spectrum(self.__spec_ptr * scale)

    def to_windows(self, window_length: float = 10, overlapping: bool = True, min_num_peaks: int = 5,
                   min_intensity: float = 1) -> Tuple[NDArray, List[MzSpectrum]]:
        """Convert the spectrum to a list of windows.

        Args:
            window_length (float, optional): Window length. Defaults to 10.
            overlapping (bool, optional): Whether the windows should overlap. Defaults to True.
            min_num_peaks (int, optional): Minimum number of peaks in a window. Defaults to 5.
            min_intensity (float, optional): Minimum intensity of a peak in a window. Defaults to 1.

        Returns:
            Tuple[NDArray, List[MzSpectrum]]: List of windows.
        """

        indices, windows = self.__spec_ptr.to_windows(window_length, overlapping, min_num_peaks, min_intensity)
        return indices, [MzSpectrum.from_py_mz_spectrum(window) for window in windows]

    def to_resolution(self, resolution: int) -> MzSpectrum:
        """Bins the spectrum's m/z values to a 
        given resolution and sums the intensities.

        Args:
            resolution (int): Negative decadic logarithm of bin size.

        Returns:
            MzSpectrum: A new `MzSpectrum` where m/z values are binned according to the given resolution.
        """
        return MzSpectrum.from_py_mz_spectrum(self.__spec_ptr.to_resolution(resolution))
    
    def filter(self, mz_min: float = 0.0, mz_max: float = 2000.0, intensity_min: float = 0.0,
               intensity_max: float = 1e9) -> MzSpectrum:
        """Filter the spectrum for a given m/z range and intensity range.

        Args:
            mz_min (float): Minimum m/z value.
            mz_max (float): Maximum m/z value.
            intensity_min (float, optional): Minimum intensity value. Defaults to 0.0.
            intensity_max (float, optional): Maximum intensity value. Defaults to 1e9.

        Returns:
            MzSpectrum: Filtered spectrum.
        """
        return MzSpectrum.from_py_mz_spectrum(
            self.__spec_ptr.filter_ranged(mz_min, mz_max, intensity_min, intensity_max))

    def vectorized(self, resolution: int = 2) -> MzSpectrumVectorized:
        """Convert the spectrum to a vectorized spectrum.

        Args:
            resolution (int, optional): Resolution. Defaults to 2.

        Returns:
            MzSpectrumVectorized: Vectorized spectrum.
        """
        return MzSpectrumVectorized.from_py_mz_spectrum_vectorized(self.__spec_ptr.vectorized(resolution))
    
    def to_jsons(self) -> str:
        """
        generates json string representation of MzSpectrum
        """
        json_dict = {}
        json_dict["mz"] = self.mz.tolist()
        json_dict["intensity"] = self.intensity.tolist()

        return json.dumps(json_dict)

    def to_centroided(self, baseline_noise_level: int = 0.0, sigma: float = 0.1, normalize: bool = True):
        """Convert the spectrum to a centroided spectrum.

        Returns:
            MzSpectrum: Centroided spectrum.
        """
        # first generate dense spectrum
        return MzSpectrum.from_py_mz_spectrum(self.__spec_ptr.to_centroided(baseline_noise_level, sigma, normalize))

    def get_spec_ptr(self) -> ims.PyMzSpectrum:
        """Get the spec_ptr.

        Returns:
            pims.PyMzSpectrum: spec_ptr.
        """
        return self.__spec_ptr

    
class MzSpectrumVectorized:
    def __init__(self, indices: NDArray[np.int32], values: NDArray[np.float64], resolution: int):
        """MzSpectrum class.

        Args:
            mz (NDArray[np.float64]): m/z.
            values (NDArray[np.float64]): Intensity.

        Raises:
            AssertionError: If the length of the mz and intensity arrays are not equal.
        """
        assert len(indices) == len(values), "The length of the mz and intensity arrays must be equal."
        self.__spec_ptr = ims.PyMzSpectrumVectorized(indices, values, resolution)

    @classmethod
    def from_py_mz_spectrum_vectorized(cls, spec: ims.PyMzSpectrumVectorized):
        """Create a MzSpectrum from a PyMzSpectrum.

        Args:
            spec (pims.PyMzSpectrum): PyMzSpectrum to create the MzSpectrum from.

        Returns:
            MzSpectrum: MzSpectrum created from the PyMzSpectrum.
        """
        instance = cls.__new__(cls)
        instance.__spec_ptr = spec
        return instance

    @property
    def resolution(self) -> float:
        """Resolution.
        
        Returns:
            float: Resolution.
        """
        return self.__spec_ptr.resolution

    @property
    def indices(self) -> NDArray[np.int32]:
        """m/z.

        Returns:
            NDArray[np.float64]: m/z.
        """
        return self.__spec_ptr.indices

    @property
    def values(self) -> NDArray[np.float64]:
        """Intensity.

        Returns:
            NDArray[np.float64]: Intensity.
        """
        return self.__spec_ptr.values

    def to_centroided(self, integrate_method: Callable = get_peak_integral) -> MzSpectrum:
        """Convert the spectrum to a centroided spectrum.

        Returns:
            MzSpectrum: Centroided spectrum.
        """
        # first generate dense spectrum
        dense_spectrum = MzSpectrumVectorized.from_py_mz_spectrum_vectorized(self.__spec_ptr.to_dense_spectrum(None))
        # find peaks in the dense spectrum and widths with scipy
        peaks, peak_info = find_peaks(dense_spectrum.values, height=0, width=(0, 0.5))
        # then get the peak integrals
        integrals = integrate_method(peaks, peak_info)
        # then create a new spectrum with the peak indices and the integrals
        return MzSpectrum.from_py_mz_spectrum(ims.PyMzSpectrum(dense_spectrum.indices[peaks]/np.power(10, dense_spectrum.resolution), integrals))

    def __repr__(self):
        return f"MzSpectrumVectorized(num_values={len(self.values)})"


class TimsSpectrum:
    def __init__(self, frame_id: int, scan: int, retention_time: float, mobility: float, ms_type: int,
                 index: NDArray[np.int32], mz: NDArray[np.float64], intensity: NDArray[np.float64]):
        """TimsSpectrum class.

        Args:
            index (NDArray[np.int32]): Index.
            mz (NDArray[np.float64]): m/z.
            intensity (NDArray[np.float64]): Intensity.

        Raises:
            AssertionError: If the length of the index, mz and intensity arrays are not equal.
        """
        assert len(index) == len(mz) == len(intensity), ("The length of the index, mz and intensity arrays must be "
                                                         "equal.")
        self.__spec_ptr = ims.PyTimsSpectrum(frame_id, scan, retention_time, mobility, ms_type, index, mz, intensity)

    @classmethod
    def from_py_tims_spectrum(cls, spec: ims.PyTimsSpectrum):
        """Create a TimsSpectrum from a PyTimsSpectrum.

        Args:
            spec (pims.PyTimsSpectrum): PyTimsSpectrum to create the TimsSpectrum from.

        Returns:
            TimsSpectrum: TimsSpectrum created from the PyTimsSpectrum.
        """
        instance = cls.__new__(cls)
        instance.__spec_ptr = spec
        return instance

    @property
    def index(self) -> NDArray[np.int32]:
        """Index.

        Returns:
            NDArray[np.int32]: Index.
        """
        return self.__spec_ptr.index

    @property
    def mz(self) -> NDArray[np.float64]:
        """m/z.

        Returns:
            NDArray[np.float64]: m/z.
        """
        return self.__spec_ptr.mz

    @property
    def intensity(self) -> NDArray[np.float64]:
        """Intensity.

        Returns:
            NDArray[np.float64]: Intensity.
        """
        return self.__spec_ptr.intensity

    @property
    def ms_type(self) -> str:
        """MS type.

        Returns:
            str: MS type.
        """
        return self.__spec_ptr.ms_type

    @property
    def mobility(self) -> float:
        """Inverse mobility.

        Returns:
            float: Inverse mobility.
        """
        return self.__spec_ptr.mobility

    @property
    def scan(self) -> int:
        """Scan.

        Returns:
            int: Scan.
        """
        return self.__spec_ptr.scan

    @property
    def retention_time(self) -> float:
        """Retention time.

        Returns:
            float: Retention time.
        """
        return self.__spec_ptr.retention_time

    @property
    def frame_id(self) -> int:
        """Frame ID.

        Returns:
            int: Frame ID.
        """
        return self.__spec_ptr.frame_id

    @property
    def mz_spectrum(self) -> MzSpectrum:
        """Get the MzSpectrum.

        Returns:
            MzSpectrum: Spectrum.
        """
        return MzSpectrum.from_py_mz_spectrum(self.__spec_ptr.mz_spectrum)

    @property
    def df(self) -> pd.DataFrame:
        """Data.
        
        Returns:
            pd.DataFrame: Data.
        """

        return pd.DataFrame({
            'frame': np.repeat(self.frame_id, len(self.index)),
            'retention_time': np.repeat(self.retention_time, len(self.index)),
            'scan': np.repeat(self.scan, len(self.index)),
            'mobility': np.repeat(self.mobility, len(self.index)),
            'tof': self.index,
            'mz': self.mz,
            'intensity': self.intensity
        })

    def __repr__(self):
        return (f"TimsSpectrum(id={self.frame_id}, retention_time={np.round(self.retention_time, 2)}, "
                f"scan={self.scan}, mobility={np.round(self.mobility, 2)}, ms_type={self.ms_type}, "
                f"num_peaks={len(self.index)})")

    def filter(self, mz_min: float = 0.0, mz_max: float = 2000.0, intensity_min: float = 0.0,
               intensity_max: float = 1e9) -> 'TimsSpectrum':
        """Filter the spectrum for a given m/z range and intensity range.

        Args:
            mz_min (float): Minimum m/z value.
            mz_max (float): Maximum m/z value.
            intensity_min (float, optional): Minimum intensity value. Defaults to 0.0.
            intensity_max (float, optional): Maximum intensity value. Defaults to 1e9.

        Returns:
            TimsSpectrum: Filtered spectrum.
        """
        return TimsSpectrum.from_py_tims_spectrum(
            self.__spec_ptr.filter_ranged(mz_min, mz_max, intensity_min, intensity_max))

    def get_spec_ptr(self) -> ims.PyTimsSpectrum:
        """Get the spec_ptr.

        Returns:
            pims.PyTimsSpectrum: spec_ptr.
        """
        return self.__spec_ptr

    def __add__(self, other):
        """Overwrite + operator for adding of spectra

        Args:
            other (TimsSpectrum): Other spectrum.

        Returns:
            TimsSpectrum: Sum of spectra
        """
        return self.from_py_tims_spectrum(self.__spec_ptr + other.__spec_ptr)



# from .mixture import GaussianMixtureModel
from .ccs.predictors import DeepPeptideIonMobilityApex, GRUCCSPredictor, load_deep_ccs_predictor
from .rt.predictors import DeepChromatographyApex, GRURetentionTimePredictor, load_deep_retention_time
from .ionization.predictors import DeepChargeStateDistribution, GRUChargeStatePredictor, load_deep_charge_state_predictor
from .utility import load_tokenizer_from_resources




import imspy_connector
ims = imspy_connector.py_constants

MASS_PROTON = ims.mass_proton()
MASS_NEUTRON = ims.mass_neutron()
MASS_ELECTRON = ims.mass_electron()
MASS_WATER = ims.mass_water()
STANDARD_TEMPERATURE = ims.standard_temperature()
STANDARD_PRESSURE = ims.standard_pressure()
ELEMENTARY_CHARGE = ims.elementary_charge()
K_BOLTZMANN = ims.k_boltzmann()
AVOGADRO_NUMBER = ims.avogadro()



import imspy_connector
ims = imspy_connector.py_amino_acids


AMINO_ACID_MASSES = ims.get_amino_acid_mono_isotopic_masses()
AMINO_ACIDS = ims.get_amino_acids()
AMINO_ACID_ATOMIC_COMPOSITIONS = ims.get_amino_acid_atomic_compositions()
